# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
from typing import Union, Iterable
from ..core import Plot, Array
from .render import render
from ..core.tools import to_bin_centers, finmin, finmax


def histogram1d(*layers: Union[Iterable, Array],
                bins: Union[int, Iterable] = 50,
                weights: Array = None,
                logx: bool = False,
                logy: bool = False,
                loglog: bool = False,
                filename: str = None,
                title: str = None,
                ymin: float = None,
                ymax: float = None,
                ax: object = None,
                **kwargs) -> Plot:
    """
    Plot a 1D histogram with arbitrary number of variables as input.
    When a vector quantity is supplied, the function will histogram the norm of
    the vectors.


    This function has an API very close to that of matplotlib's ``hist`` function.
    For the documentation of any parameters that are not listed below, see
    https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.hist.html.

    :param layers: Dicts or Arrays representing the quantities to be mapped onto the
        colormap of the generated image.

    :param logx: If ``True``, use logarithmic scaling on the horizontal axis.
        Default is ``False``.

    :param logy: If ``True``, use logarithmic scaling on the vertical axis.
        Default is ``False``.

    :param loglog: If ``True``, use logarithmic scaling on the horizontal and
        vertical axes. Default is ``False``.

    :param filename: If specified, the returned figure is also saved to file.
        Default is ``None``.

    :param title: The title of the figure. Default is ``None``.

    :param ymin: Minimum value for the vertical axis. Default is ``None``.

    :param ymax: Maximum value for the vertical axis. Default is ``None``.

    :param ax: A matplotlib axes inside which the figure will be plotted.
        Default is ``None``, in which case some new axes a created.
    """
    if loglog:
        logx = logy = True

    figure = render(logx=logx, logy=logy, ax=ax)

    for layer in layers:
        if isinstance(layer, dict):
            params = {}
            extra_args = {}
            for key, param in layer.items():
                if key in ["data", "bins", "weights"]:
                    params[key] = param
                else:
                    extra_args[key] = param
            for key, arg in {'bins': bins, 'weights': weights}.items():
                if key not in params:
                    params[key] = arg
        else:
            params = {'data': layer, 'bins': bins, 'weights': weights}
            extra_args = kwargs

        xvals = params['data'].norm.values
        if params['weights'] is not None:
            params['weights'] = params['weights'].norm.values

        # Construct some bin edges
        if isinstance(params['bins'], int):
            xmin = finmin(xvals)
            xmax = finmax(xvals)
            if logx:
                xedges = np.logspace(np.log10(xmin), np.log10(xmax), params['bins'] + 1)
            else:
                xedges = np.linspace(xmin, xmax, params['bins'] + 1)
        else:
            xedges = params['bins']

        ydata, _, _ = figure["ax"].hist(xvals,
                                        bins=xedges,
                                        weights=params['weights'],
                                        **extra_args)

        figure["ax"].set_xlabel(params['data'].label)

    figure["ax"].set_ylim(ymin, ymax)
    return Plot(x=to_bin_centers(xedges),
                y=ydata,
                fig=figure["fig"],
                ax=figure["ax"],
                filename=filename)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)
import numpy as np
from .. import config
from ..io import Loader
from .datagroup import Datagroup
from .tools import bytes_to_human_readable


class Dataset:
    def __init__(self, nout=None, scale=None, path=""):
        self.groups = {}
        self.meta = {}
        self.loader = None
        if scale is None:
            scale = config.parameters["scale"]
        if nout is not None:
            self.loader = Loader(nout=nout, scale=scale, path=path)
            self.meta.update(self.loader.load_metadata())

    def __iter__(self):
        return self.groups.__iter__()

    def __len__(self):
        return self.groups.__len__()

    def __getitem__(self, key):
        return self.groups[key]

    def __setitem__(self, key, value):
        if not isinstance(value, Datagroup):
            raise TypeError("Only objects of type Datagroup can be inserted into a "
                            "Dataset.")
        self.groups.__setitem__(key, value)
        value.name = key
        value.parent = self

    def __delitem__(self, key):
        return self.groups.__delitem__(key)

    def __repr__(self):
        return str(self)

    def __str__(self):
        output = "Dataset: "
        if "infile" in self.meta:
            output += "{}: ".format(self.meta["infile"])
        output += "{}\n".format(self.print_size())
        for key, item in self.items():
            output += str(item) + "\n"
        return output

    def keys(self):
        return self.groups.keys()

    def items(self):
        return self.groups.items()

    def values(self):
        return self.groups.values()

    def load(self, *args, **kwargs):
        groups = self.loader.load(*args, meta=self.meta, **kwargs)
        for name, group in groups.items():
            self[name] = group
        config.additional_variables(self)
        return self

    def nbytes(self):
        return np.sum([item.nbytes() for item in self.groups.values()])

    def print_size(self):
        return bytes_to_human_readable(self.nbytes())



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
from numba import njit, prange


@njit(parallel=True)
def evaluate_on_grid(cell_positions_in_new_basis, cell_positions_in_original_basis,
                     cell_values, cell_sizes, grid_lower_edge_in_new_basis,
                     grid_spacing_in_new_basis, grid_positions_in_original_basis, ndim):

    nz, ny, nx = grid_positions_in_original_basis.shape[:3]
    diagonal = np.sqrt(ndim)
    out = np.full(shape=(cell_values.shape[0], nz, ny, nx),
                  fill_value=np.nan,
                  dtype=np.float64)

    ncells = cell_positions_in_new_basis.shape[0]
    for n in prange(ncells):

        half_size = cell_sizes[n] * diagonal
        ix1 = max(
            int(((cell_positions_in_new_basis[n, 0] - half_size) -
                 grid_lower_edge_in_new_basis[0]) / grid_spacing_in_new_basis[0]), 0)
        ix2 = min(
            int(((cell_positions_in_new_basis[n, 0] + half_size) -
                 grid_lower_edge_in_new_basis[0]) / grid_spacing_in_new_basis[0]) + 1,
            nx)
        iy1 = max(
            int(((cell_positions_in_new_basis[n, 1] - half_size) -
                 grid_lower_edge_in_new_basis[1]) / grid_spacing_in_new_basis[1]), 0)
        iy2 = min(
            int(((cell_positions_in_new_basis[n, 1] + half_size) -
                 grid_lower_edge_in_new_basis[1]) / grid_spacing_in_new_basis[1]) + 1,
            ny)
        iz1 = max(
            int(((cell_positions_in_new_basis[n, 2] - half_size) -
                 grid_lower_edge_in_new_basis[2]) / grid_spacing_in_new_basis[2]), 0)
        iz2 = min(
            int(((cell_positions_in_new_basis[n, 2] + half_size) -
                 grid_lower_edge_in_new_basis[2]) / grid_spacing_in_new_basis[2]) + 1,
            nz)

        for k in range(iz1, iz2):
            for j in range(iy1, iy2):
                for i in range(ix1, ix2):
                    dist = np.abs(grid_positions_in_original_basis[k, j, i, :] -
                                  cell_positions_in_original_basis[n, :])
                    if np.all(dist <= cell_sizes[n]):
                        out[:, k, j, i] = cell_values[:, n]

    return out


@njit(parallel=True)
def hist2d(x, y, values, xmin, xmax, nx, ymin, ymax, ny):

    out = np.zeros(shape=(values.shape[0], ny, nx), dtype=np.float64)
    counts = np.zeros(shape=(ny, nx), dtype=np.int64)
    dx = (xmax - xmin) / nx
    dy = (ymax - ymin) / ny

    for i in prange(len(x)):
        indx = int((x[i] - xmin) / dx)
        indy = int((y[i] - ymin) / dy)
        if (indx >= 0) and (indx < nx) and (indy >= 0) and (indy < ny):
            out[:, indy, indx] += values[:, i]
            counts[indy, indx] += 1

    return out, counts



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
from pint.quantity import Quantity
from typing import Union
from ..core import Array, Plot
from .render import render
from ..core.tools import to_bin_centers, finmin, finmax
from .parser import parse_layer
from .utils import hist2d


def _parse_limit(limit, x, logx, reduction):
    autox = False
    if limit is None:
        if reduction == "min":
            limit = finmin(x.values)
        elif reduction == "max":
            limit = finmax(x.values)
        else:
            raise RuntimeError(
                f"_parse_limit: unknown reduction operation {reduction}.")
        autox = True
    else:
        if isinstance(limit, Quantity):
            limit = limit.to(x.unit.units).magnitude
        if logx:
            limit = np.log10(limit)
    return limit, autox


def histogram2d(x: Array,
                y: Array,
                *layers,
                mode: str = None,
                logx: bool = False,
                logy: bool = False,
                loglog: bool = False,
                norm: str = None,
                filename: str = None,
                resolution: Union[int, dict] = 256,
                operation: str = "sum",
                title: str = None,
                xmin: float = None,
                xmax: float = None,
                ymin: float = None,
                ymax: float = None,
                vmin: float = None,
                vmax: float = None,
                plot: bool = True,
                ax: object = None,
                **kwargs) -> Plot:
    """
    Plot a 2D histogram with two variables as input.
    When a vector quantity is supplied, the function will histogram the norm of
    the vectors.


    :param x: Horizontal Array to be histogrammed.

    :param y: Vertical Array to be histogrammed.

    :param layers: Dicts or Arrays representing the quantities to be mapped onto the
        colormap of the generated image.

    :param mode: The rendering mode for the histogram. Possible choices are
        ``'image'``, ``'contourf'``, ``'contour'``, and ``'scatter'``. Default is
        ``None``, which selects the ``render_mode`` set in the user configuration
        file (``'image'`` by default).

    :param logx: If ``True``, use logarithmic scaling on the horizontal axis.
        Default is ``False``.

    :param logy: If ``True``, use logarithmic scaling on the vertical axis.
        Default is ``False``.

    :param loglog: If ``True``, use logarithmic scaling on the horizontal and
        vertical axes. Default is ``False``.

    :param norm: The colormap normalization. Possible values are ``'linear'`` and
        ``'log'``. Default is ``None`` (= ``'linear'``).

    :param filename: If specified, the returned figure is also saved to file.
        Default is ``None``.

    :param resolution: Resolution of the generated map. This can either be an
        integer or a dict. In the case of an integer, it represents the number of
        pixels used for the horizontal and vertical dimensions. For a dictionary,
        the following syntax should be used: ``resolution={'x': 128, 'y': 192}``.
        Default is ``256``.

    :param operation: The operation to apply inside the bins of the histogram.
        Possible values are ``'sum'`` and ``'mean'``. Default is ``'sum'``.

    :param title: The title of the figure. Default is ``None``.

    :param xmin: Minimum value for the horizontal axis. Default is ``None``.

    :param xmax: Maximum value for the horizontal axis. Default is ``None``.

    :param ymin: Minimum value for the vertical axis. Default is ``None``.

    :param ymax: Maximum value for the vertical axis. Default is ``None``.

    :param vmin: Minimum value for colorbar range. Default is ``None``.

    :param vmax: Maximum value for colorbar range. Default is ``None``.

    :param plot: Make a plot if ``True``. If not, just return the ``Plot`` object
        containing the data that would be used to generate the plot.
        Default is ``True``.

    :param ax: A matplotlib axes inside which the figure will be plotted.
        Default is ``None``, in which case some new axes a created.
    """

    x = x.norm
    y = y.norm

    if loglog:
        logx = logy = True
    if logx:
        x = np.log10(x)
    if logy:
        y = np.log10(y)

    nx = resolution
    ny = resolution

    xmin, autoxmin = _parse_limit(xmin, x, logx, "min")
    xmax, autoxmax = _parse_limit(xmax, x, logx, "max")
    ymin, autoymin = _parse_limit(ymin, y, logy, "min")
    ymax, autoymax = _parse_limit(ymax, y, logy, "max")

    # Protect against empty plots if xmin==xmax or ymin==ymax
    if xmin == xmax:
        if xmin == 0.0:
            xmin = -0.1
            xmax = 0.1
        else:
            xmin = xmin - 0.05 * abs(xmin)
            xmax = xmax + 0.05 * abs(xmax)
    if ymin == ymax:
        if ymin == 0.0:
            ymin = -0.1
            ymax = 0.1
        else:
            ymin = ymin - 0.05 * abs(ymin)
            ymax = ymax + 0.05 * abs(ymax)

    dx = xmax - xmin
    dy = ymax - ymin
    if autoxmin:
        xmin = xmin - 0.05 * dx
    if autoxmax:
        xmax = xmax + 0.05 * dx
    if autoymin:
        ymin = ymin - 0.05 * dy
    if autoymax:
        ymax = ymax + 0.05 * dy

    # Construct some bin edges and centers
    if logx:
        xedges = np.logspace(xmin, xmax, nx + 1)
    else:
        xedges = np.linspace(xmin, xmax, nx + 1)
    if logy:
        yedges = np.logspace(ymin, ymax, ny + 1)
    else:
        yedges = np.linspace(ymin, ymax, ny + 1)
    xcenters = to_bin_centers(xedges)
    ycenters = to_bin_centers(yedges)

    to_render = []
    to_process = []
    operations = []

    xvals = x.values
    yvals = y.values

    # If no layers are defined, make a layer for counting cells
    if len(layers) == 0:
        layers = [Array(values=np.ones_like(xvals), name="counts")]

    for layer in layers:
        data, settings, params = parse_layer(layer=layer,
                                             mode=mode,
                                             norm=norm,
                                             vmin=vmin,
                                             vmax=vmax,
                                             operation=operation,
                                             **kwargs)
        to_process.append(data.norm.values)
        to_render.append({
            "mode": settings["mode"],
            "params": params,
            "unit": data.unit.units,
            "name": data.name
        })
        operations.append(settings["operation"])

    # Send to numba histogramming
    binned, counts = hist2d(x=xvals,
                            y=yvals,
                            values=np.array(to_process),
                            xmin=xmin,
                            xmax=xmax,
                            nx=nx,
                            ymin=ymin,
                            ymax=ymax,
                            ny=ny)

    mask = counts == 0
    for ind in range(len(to_process)):
        if operations[ind] == "mean":
            with np.errstate(invalid="ignore"):
                binned[ind, ...] /= counts
        to_render[ind]["data"] = np.ma.masked_where(mask, binned[ind, ...])

    to_return = {
        "x": xcenters,
        "y": ycenters,
        "layers": to_render,
        "filename": filename
    }

    if plot:
        figure = render(x=xcenters,
                        y=ycenters,
                        data=to_render,
                        logx=logx,
                        logy=logy,
                        ax=ax)
        figure["ax"].set_xlabel(x.label)
        figure["ax"].set_ylabel(y.label)
        to_return.update({"fig": figure["fig"], "ax": figure["ax"]})

    return Plot(**to_return)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)
import numpy as np
from .array import Array
from .tools import bytes_to_human_readable


class Datagroup:
    def __init__(self, data=None, parent=None):
        self._container = {}
        self._parent = parent
        self._name = ""
        self.shape = None
        if data is not None:
            for key, array in data.items():
                self[key] = array

    def __iter__(self):
        return self._container.__iter__()

    def __len__(self):
        return self._container.__len__()

    def __getitem__(self, key):
        if isinstance(key, str):
            return self._container[key]
        else:
            if isinstance(key, int):
                key = slice(key, key + 1, 1)
            d = self.__class__()
            for name, val in self.items():
                d[name] = val[key]
            return d

    def __setitem__(self, key, value):
        if len(value.shape) > 0:
            shape = value.shape[0]
        else:
            shape = 1
        if self.shape is None:
            self.shape = shape
        else:
            if self.shape != shape:
                raise RuntimeError(
                    "Size mismatch on element insertion. Item "
                    "shape is {} while container accepts shape {}.".format(
                        shape, self.shape))
        if isinstance(value, Array):
            value.name = key
            value.parent = self
            self._container[key] = value
        else:
            self._container[key] = Array(values=value, name=key, parent=self)

    def __delitem__(self, key):
        return self._container.__delitem__(key)

    def __repr__(self):
        return str(self)

    def __str__(self):
        output = "Datagroup: {} {}\n".format(self.name, self.print_size())
        for key, item in self.items():
            output += str(item) + "\n"
        return output

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent_):
        self._parent = parent_

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name_):
        self._name = name_

    def keys(self):
        return self._container.keys()

    def items(self):
        return self._container.items()

    def values(self):
        return self._container.values()

    def set_scale(self, scale):
        for key in ["x", "y", "z", "dx"]:
            if key in self:
                self[key].to(scale)

    def nbytes(self):
        return np.sum([item._array.nbytes for item in self.values()])

    def print_size(self):
        return bytes_to_human_readable(self.nbytes())



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

# flake8: noqa

from .histogram1d import histogram1d
from .histogram2d import histogram2d
from .map import map
from .render import render
from .plane import plane
from .plot import plot
from .scatter import scatter



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

from .. import config
from ..core import Array
from contextlib import redirect_stderr
import io
import matplotlib.pyplot as plt
from matplotlib.cm import ScalarMappable
from matplotlib.collections import PatchCollection
import numpy as np
from pint.quantity import Quantity


def _add_colorbar(obj, ax, cax=None, label=None):
    """
    Add colorbar to plot for a given artist.
    """
    cb = plt.colorbar(obj, ax=ax, cax=cax)
    cb.set_label(label)
    cb.ax.yaxis.set_label_coords(-1.05, 0.5)


def quiver(ax,
           x,
           y,
           z,
           cbar=False,
           cblabel=None,
           density=1,
           color="w",
           zorder=2,
           **kwargs):
    """
    Wrapper around Matplotlib's quiver plot.

    We have a special argument `density` which aims to control the density of
    arrows in a similar manner to the density of streamlines in the streamplot.
    Matplotlib's own way of controlling the density of arrows is not so
    obvious.
    """
    default_args = {"angles": "xy", "pivot": "mid", "zorder": zorder}
    default_args.update(kwargs)

    skips = np.around(np.array(z.shape) * 4.0 / 128.0 / density).astype(np.int)
    skip = (slice(None, None, skips[0]), slice(None, None, skips[1]))

    args = [x[skip[0]], y[skip[1]], z[..., 0][skip], z[..., 1][skip]]
    if isinstance(color, str):
        default_args["color"] = color
    else:
        args.append(z[..., 2][skip])

    out = ax.quiver(*args, **default_args)
    if cbar and not isinstance(color, str):
        _add_colorbar(obj=out, ax=ax, label=cblabel)
    return out


def pcolormesh(ax, x, y, z, cbar=False, cblabel=None, zorder=1, **kwargs):
    """
    Wrapper around Matplotlib's pcolormesh plot.
    """
    default_args = {
        "shading": "nearest",
        "zorder": zorder,
        "cmap": config.parameters["cmap"]
    }
    default_args.update(kwargs)
    out = ax.pcolormesh(x, y, z, **default_args)
    if cbar:
        _add_colorbar(obj=out, ax=ax, label=cblabel)
    return out


def contour(ax,
            x,
            y,
            z,
            cbar=False,
            cblabel=None,
            labels=True,
            zorder=2,
            fmt='%1.3f',
            **kwargs):
    """
    Wrapper around Matplotlib's contour plot.

    We add a small convenience argument `labels` that allows to simply add
    clabels with a simple `labels=True`.
    """
    cs = ax.contour(x, y, z, zorder=zorder, **kwargs)
    if labels:
        ax.clabel(cs, cs.levels, inline=1, fontsize=10, fmt=fmt)
    return cs


def contourf(ax, x, y, z, cbar=False, cblabel=None, zorder=1, **kwargs):
    """
    Wrapper around Matplotlib's contourf plot.
    """
    if "cmap" not in kwargs:
        kwargs["cmap"] = config.parameters["cmap"]
    out = ax.contourf(x, y, z, **kwargs)
    if cbar:
        _add_colorbar(obj=out, ax=ax, label=cblabel)
    return out


def streamplot(ax, x, y, z, cbar=False, cblabel=None, color='w', zorder=2, **kwargs):
    """
    Wrapper around Matplotlib's streamplot plot.
    """
    default_args = {"color": "w", "zorder": zorder, "cmap": config.parameters["cmap"]}
    default_args.update(kwargs)
    if isinstance(color, str):
        default_args["color"] = color
    else:
        default_args["color"] = z[..., 2]
        if default_args["norm"].vmin is None:
            default_args["norm"].vmin = default_args["color"].min()
        if default_args["norm"].vmax is None:
            default_args["norm"].vmax = default_args["color"].max()
    out = ax.streamplot(x, y, z[..., 0], z[..., 1], **default_args)
    if cbar and not isinstance(color, str):
        _add_colorbar(obj=out.lines, ax=ax, label=cblabel)
    return out


def scatter(ax, x, y, z, cbar=False, cblabel=None, zorder=2, **kwargs):
    """
    Wrapper around Matplotlib's scatter plot.
    If a point size has a unit, use PatchCollection instead of scatter.
    We use the scatter API for the arguments.
    If PatchCollection is used, we convert the scatter args to the
    PatchCollection syntax (e.g. "c" -> "color").
    """
    default_args = {"c": "C0", "edgecolors": "k", "zorder": zorder}
    default_args.update(kwargs)
    use_patchcollection = False
    need_cbar = False
    if "s" in default_args:
        if isinstance(default_args["s"], Array):
            default_args["s"] = default_args["s"].norm.values
            use_patchcollection = True
        if isinstance(default_args["s"], Quantity):
            default_args["s"] = np.full_like(x, default_args["s"].magnitude)
            use_patchcollection = True
    if use_patchcollection:
        array = None
        norm = None
        if "c" in default_args:
            need_cbar = not isinstance(default_args["c"], str)
            if "facecolors" not in default_args:
                if isinstance(default_args["c"], str):
                    default_args["facecolors"] = default_args["c"]
                else:
                    array = default_args["c"]
            del default_args["c"]
        if "norm" in default_args:
            norm = default_args["norm"]
            del default_args["norm"]
        try:
            iter(default_args["s"])
        except TypeError:
            default_args["s"] = [default_args["s"]] * len(x)
        patches = [
            plt.Circle([x_, y_], s) for x_, y_, s in zip(x, y, default_args["s"])
        ]
        del default_args["s"]
        out = ax.add_collection(PatchCollection(patches, **default_args))
        if array is not None:
            out.set_array(array)
        if norm is not None:
            out.set_norm(norm)
    else:
        need_cbar = not isinstance(default_args.get("c", ""), str)
        out = ax.scatter(x, y, **default_args)
    if cbar and need_cbar:
        _add_colorbar(obj=out, ax=ax, label=cblabel)
    return out


def line_integral_convolution(ax,
                              x,
                              y,
                              z,
                              cbar=False,
                              cblabel=None,
                              length=None,
                              color=None,
                              **kwargs):
    """
    Wrapper that plots a line integral convolution of a vector field.
    Uses alpha blending to merge the LIC with a user-requested color.
    """
    from lic import lic

    # Compute line integral convolution
    if length is None:
        length = int(max(z.shape[:-1]) * 15 / 128)
    with redirect_stderr(io.StringIO()) as _:
        lic_res = lic(z[..., 1], z[..., 0], length=length)

    if color is not None:
        plot_args = {**kwargs}
        base_args = {"norm": None, "cmap": None}
        for key in base_args:
            if key in plot_args:
                base_args[key] = plot_args[key]
                del plot_args[key]
        scalar_map = ScalarMappable(**base_args)
        base_rgba = scalar_map.to_rgba(z[..., 2])
        base_alpha = 1.0

        lim = (.2, .5)
        lic_data_clip = np.clip(lic_res, lim[0], lim[1])
        lic_rgba = ScalarMappable(norm=None, cmap="binary").to_rgba(lic_data_clip)
        lic_data_clip_rescale = (lic_data_clip - lim[0]) / (lim[1] - lim[0])
        lic_alpha = lic_data_clip_rescale.reshape(lic_data_clip_rescale.shape +
                                                  (1, )) * 0.3
        # Perform alpha blending manually
        rgba = lic_rgba * lic_alpha + (base_alpha * (1.0 - lic_alpha)) * base_rgba
        out = ax.imshow(rgba,
                        extent=[
                            0.5 * (3 * x[0] - x[1]), 0.5 * (3 * x[-1] - x[-2]),
                            0.5 * (3 * y[0] - y[1]), 0.5 * (3 * y[-1] - y[-2])
                        ],
                        origin='lower',
                        zorder=1)
        # Add the colorbar using the ScalarMappable
        scalar_map.set_array(z[..., 2])
        if cbar:
            _add_colorbar(obj=scalar_map, ax=ax, label=cblabel)
        return out
    else:
        return pcolormesh(ax=ax, x=x, y=y, z=lic_res, cbar=cbar, cblabel="", **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
import os
from . import utils
from .. import config
from ..core import Datagroup
from .amr import AmrReader
from .grav import GravReader
from .hydro import HydroReader
from .part import PartReader
from .rt import RtReader
from .sink import SinkReader
from .reader import ReaderKind


class Loader:
    def __init__(self, nout, scale, path):
        # Generate directory name from output number
        self.nout = nout
        self.path = path
        self.scale = scale
        self.infile = utils.generate_fname(nout, path)
        self.readers = {
            "amr": AmrReader(),
            "hydro": HydroReader(),
            "grav": GravReader(),
            "part": PartReader(),
            "rt": RtReader(),
            "sink": SinkReader()
        }

    def load_metadata(self):

        # Read info file and create info dictionary
        infofile = os.path.join(self.infile,
                                "info_" + self.infile.split("_")[-1] + ".txt")
        meta = utils.read_parameter_file(fname=infofile)
        # Add additional information
        meta["infofile"] = infofile
        meta["scale"] = self.scale
        meta["infile"] = self.infile
        meta["nout"] = self.nout
        meta["path"] = self.path
        meta["time"] *= config.get_unit("time", meta["unit_d"], meta["unit_l"],
                                        meta["unit_t"], meta["scale"])
        meta["ncells"] = 0
        meta["nparticles"] = 0
        return meta

    def load(self, select=None, cpu_list=None, meta=None):

        out = {}
        groups = list(self.readers.keys())

        _select = {group: {} for group in self.readers}
        if isinstance(select, dict):
            for key in select:
                if key not in self.readers:
                    print("Warning: {} found in select is not a valid "
                          "Datagroup.".format(key))
                else:
                    _select[key] = select[key]
        elif isinstance(select, str):
            for key in _select:
                if key != select:
                    _select[key] = False
        elif isinstance(select, list) or isinstance(select, tuple):
            for key in _select:
                if key not in select:
                    _select[key] = False

        # Take into account user specified lmax
        meta["lmax"] = meta["levelmax"]
        if "amr" in _select:
            if _select["amr"]:
                if "level" in _select["amr"]:
                    meta["lmax"] = utils.find_max_amr_level(levelmax=meta["levelmax"],
                                                            select=_select["amr"])

        # Initialize readers
        readers = {}
        for group in groups:
            loaded_on_init = self.readers[group].initialize(meta=meta,
                                                            select=_select[group])
            if loaded_on_init is not None:
                out[group] = loaded_on_init
            if self.readers[group].initialized:
                readers[group] = self.readers[group]

        # Take into account user specified cpu list
        if cpu_list is None:
            cpu_list = self.readers["amr"].cpu_list if self.readers[
                "amr"].cpu_list is not None else range(1, meta["ncpu"] + 1)

        print("Processing {} files in {}".format(len(cpu_list), meta["infile"]))

        # Allocate work arrays
        twotondim = 2**meta["ndim"]
        for reader in readers.values():
            reader.allocate_buffers(ngridmax=meta["ngridmax"], twotondim=twotondim)

        iprog = 1
        istep = 10
        npieces = 0

        # byte, integer, double, line, string, quad, long
        null_offsets = {key: 0 for key in "bidnsql"}

        # Loop over the cpus and read the AMR and HYDRO files in binary format
        for cpu_ind, cpu_num in enumerate(cpu_list):

            # Print progress
            percentage = int(float(cpu_ind) * 100.0 / float(len(cpu_list)))
            if percentage >= iprog * istep:
                print("{:>3d}% : read {:>10d} cells, {:>10d} particles".format(
                    percentage, meta["ncells"], meta["nparticles"]))
                iprog += 1

            # Read binary files
            for group, reader in readers.items():
                fname = utils.generate_fname(meta["nout"],
                                             meta["path"],
                                             ftype=group,
                                             cpuid=cpu_num)
                with open(fname, mode='rb') as f:
                    reader.bytes = f.read()

            # Read file headers
            for reader in readers.values():
                reader.offsets.update(null_offsets)
                reader.read_header(meta)

            # Loop over levels
            for ilevel in range(meta["lmax"]):

                for reader in readers.values():
                    reader.read_level_header(ilevel, twotondim)

                # Loop over domains
                for domain in range(readers["amr"].meta["nboundary"] + meta["ncpu"]):

                    ncache = readers["amr"].meta["ngridlevel"][domain, ilevel]

                    for reader in readers.values():
                        reader.read_domain_header()

                    if ncache > 0:

                        if domain == cpu_num - 1:

                            for reader in readers.values():
                                reader.read_cacheline_header(ncache, meta["ndim"])

                            for ind in range(twotondim):

                                # Read variables in cells
                                for reader in readers.values():
                                    reader.read_variables(ncache, ind, ilevel,
                                                          cpu_num - 1, meta)

                            # Apply selection criteria: select only leaf cells and
                            # add any criteria requested by the user via select.
                            conditions = {}
                            for group, reader in readers.items():
                                conditions.update(
                                    reader.make_conditions(_select[group], ncache))
                            # Combine all selection criteria together with AND
                            # operation by using a product on bools
                            sel = np.where(
                                np.prod(np.array(list(conditions.values())), axis=0))

                            # Count the number of cells
                            ncells = np.shape(sel)[1]
                            if ncells > 0:
                                meta["ncells"] += ncells
                                npieces += 1
                                # Add the cells in the pieces dictionaries
                                for reader in readers.values():
                                    if reader.kind == ReaderKind.AMR:
                                        for item in reader.variables.values():
                                            if item["read"]:
                                                item["pieces"][npieces] = item[
                                                    "buffer"][sel]

                            # Increment offsets with remainder of the file
                            for reader in readers.values():
                                reader.read_footer(ncache, twotondim)

                        else:

                            for reader in readers.values():
                                reader.step_over(ncache, twotondim, meta["ndim"])

        # Merge all the data pieces into the Arrays
        for group, reader in readers.items():
            out[group] = Datagroup()
            for key, item in reader.variables.items():
                if item["read"] and len(item["pieces"]) > 0:
                    out[group][key] = np.concatenate(list(item["pieces"].values()))
            # If vector quantities are found, make them into vector Arrays
            utils.make_vector_arrays(out[group], ndim=meta["ndim"])

        print("Loaded: {} cells, {} particles.".format(meta["ncells"],
                                                       meta["nparticles"]))

        return out



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

# flake8: noqa

from .config import config, units
from .plot import histogram1d, histogram2d, plane, scatter, map, plot
from .core import Array, Datagroup, Dataset, Plot



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)
import numpy as np
import os
from .reader import Reader, ReaderKind
from . import utils


class HydroReader(Reader):
    def __init__(self):
        super().__init__(kind=ReaderKind.AMR)

    def initialize(self, meta, select):
        self.initialized = False
        if select is False:
            return

        # Read the number of variables from the hydro_file_descriptor.txt
        # and select the ones to be read if specified by user
        fname = os.path.join(meta["infile"], "hydro_file_descriptor.txt")
        try:
            desc_from_file = np.loadtxt(fname, dtype=str, delimiter=",")
        except IOError:
            return

        descriptor = {
            desc_from_file[i, 1].strip(): desc_from_file[i, 2].strip()
            for i in range(len(desc_from_file))
        }

        self.descriptor_to_variables(descriptor=descriptor, meta=meta, select=select)
        self.initialized = True

    def read_header(self, info):
        # hydro gamma
        self.offsets["i"] += 5
        self.offsets["n"] += 5
        [info["gamma"]] = utils.read_binary_data(fmt="d",
                                                 content=self.bytes,
                                                 offsets=self.offsets)

    def read_domain_header(self):
        self.offsets['n'] += 2
        self.offsets['i'] += 2



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)
"""
Define default values so that you don't have to specify them every time.
"""
import math
from pint import UnitRegistry

ureg = UnitRegistry(system="cgs")

parameters = {
    'scale': 'au',
    'path': None,
    'select': None,
    'cmap': 'viridis',
    'render_mode': 'pcolormesh'
}


def get_unit(string, ud, ul, ut, scale):

    density = ud * (ureg.g / (ureg.cm**3))
    velocity = (ul / ut) * (ureg.cm / ureg.s)
    magnetic_field = math.sqrt(4.0 * math.pi * ud * (ul / ut)**2) * ureg.G
    momentum = density * velocity
    acceleration = (ul / ut**2) * (ureg.cm / (ureg.s**2))
    energy = ud * ((ul / ut)**2) * (ureg.erg / (ureg.cm**3))
    time = ut * ureg.s
    length = ul * ureg.cm
    mass = density * (length**3)

    scaling = length
    if scale is not None:
        scale = ureg(scale)
        scaling = (length.to(scale) / scale).magnitude * scale

    ramses_units = {
        'density': density,
        'velocity': velocity,
        'velocity_x': velocity,
        'velocity_y': velocity,
        'velocity_z': velocity,
        'momentum': momentum,
        'momentum_x': momentum,
        'momentum_y': momentum,
        'momentum_z': momentum,
        'B_left': magnetic_field,
        'B_left_x': magnetic_field,
        'B_left_y': magnetic_field,
        'B_left_z': magnetic_field,
        'B_right': magnetic_field,
        'B_right_x': magnetic_field,
        'B_right_y': magnetic_field,
        'B_right_z': magnetic_field,
        'B_field': magnetic_field,
        'B_field_x': magnetic_field,
        'B_field_y': magnetic_field,
        'B_field_z': magnetic_field,
        'B_x_left': magnetic_field,
        'B_y_left': magnetic_field,
        'B_z_left': magnetic_field,
        'B_x_right': magnetic_field,
        'B_y_right': magnetic_field,
        'B_z_right': magnetic_field,
        'grav_acceleration': acceleration,
        'grav_acceleration_x': acceleration,
        'grav_acceleration_y': acceleration,
        'grav_acceleration_z': acceleration,
        'thermal_pressure': energy,
        'pressure': energy,
        'radiative_energy': energy,
        'radiative_energy_1': energy,
        'temperature': 1.0 * ureg.K,
        'time': time,
        'x': scaling,
        'y': scaling,
        'z': scaling,
        'xyz_x': scaling,
        'xyz_y': scaling,
        'xyz_z': scaling,
        'position_x': scaling,
        'position_y': scaling,
        'position_z': scaling,
        'dx': scaling,
        'mass': mass
    }

    if string in ramses_units:
        return ramses_units[string]
    else:
        return 1.0 * ureg.dimensionless


def additional_units():
    """
    Define additional useful ureg and constants
    """
    ureg.define('solar_mass = 1.9889e+33 * g = msun')
    ureg.define('radiation_constant = 7.56591469318689378e-015 * erg / cm^3 / K^4 = ar')


def additional_variables(data):
    """
    Here are some additional variables that are to be computed every time data
    is loaded.

    It is recommended to place your variables in a `try/except` block, which
    will prevent errors if the variables are not found, for instance when
    loading data from a different simulation.
    """

    # Magnetic field
    try:
        data['hydro']['B_field'] = 0.5 * (data['hydro']['B_left'] +
                                          data['hydro']['B_right'])
    except KeyError:
        pass

    # Mass
    try:
        data['hydro']['mass'] = data['hydro']['density'] * data['amr']['dx']**3
        data['hydro']['mass'].to('msun')
    except KeyError:
        pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)
import numpy as np
from .hilbert import hilbert_cpu_list
from .reader import Reader, ReaderKind
from .. import config
from . import utils


class AmrReader(Reader):
    def __init__(self):
        super().__init__(kind=ReaderKind.AMR)
        self.cpu_list = None

    def initialize(self, meta, select):
        self.initialized = False

        descriptor = {"level": "i", "cpu": "i", "dx": "d"}
        descriptor.update({"xyz_{}".format(c): "d" for c in "xyz"[:meta["ndim"]]})

        self.descriptor_to_variables(descriptor=descriptor, meta=meta, select=select)

        self.cpu_list = hilbert_cpu_list(meta=meta,
                                         scaling=config.get_unit(
                                             "z", meta["unit_d"], meta["unit_l"],
                                             meta["unit_t"], meta["scale"]),
                                         select=select,
                                         infofile=meta["infofile"])
        self.initialized = True

    def allocate_buffers(self, ngridmax, twotondim):
        super().allocate_buffers(ngridmax, twotondim)
        self.xcent = np.zeros([8, 3], dtype=np.float64)
        self.xg = np.zeros([ngridmax, 3], dtype=np.float64)
        self.son = np.zeros([ngridmax, twotondim], dtype=np.int32)
        self.ref = np.zeros([ngridmax, twotondim], dtype=np.bool)

    def read_header(self, info):
        # nx,ny,nz
        self.offsets["i"] += 2
        self.offsets["n"] += 2
        [nx, ny, nz] = utils.read_binary_data(fmt="3i",
                                              content=self.bytes,
                                              offsets=self.offsets)
        ncoarse = nx * ny * nz
        self.meta["xbound"] = [
            float(int(nx / 2)),
            float(int(ny / 2)),
            float(int(nz / 2))
        ]

        # nboundary
        self.offsets["i"] += 2
        self.offsets["n"] += 2
        [self.meta["nboundary"]] = utils.read_binary_data(fmt="i",
                                                          content=self.bytes,
                                                          offsets=self.offsets)
        self.meta["ngridlevel"] = np.zeros(
            [info["ncpu"] + self.meta["nboundary"], info["levelmax"]], dtype=np.int32)

        # noutput
        self.offsets["i"] += 1
        self.offsets["n"] += 2
        self.offsets["d"] += 1
        [noutput] = utils.read_binary_data(fmt="i",
                                           content=self.bytes,
                                           offsets=self.offsets)
        # dtold, dtnew
        self.offsets["i"] += 2
        self.offsets["n"] += 3
        self.offsets["d"] += 1 + 2 * noutput
        info["dtold"] = np.array(
            utils.read_binary_data(fmt="{}d".format(info["levelmax"]),
                                   content=self.bytes,
                                   offsets=self.offsets))
        info["dtnew"] = np.array(
            utils.read_binary_data(fmt="{}d".format(info["levelmax"]),
                                   content=self.bytes,
                                   offsets=self.offsets))

        # Read the number of grids
        self.offsets["i"] += 2 + (2 * info["ncpu"] * info["levelmax"])
        self.offsets["n"] += 7
        self.offsets["d"] += 16
        self.meta["ngridlevel"][:info["ncpu"], :] = np.array(
            utils.read_binary_data(fmt="{}i".format(info["ncpu"] * info["levelmax"]),
                                   content=self.bytes,
                                   offsets=self.offsets)).reshape(
                                       info["levelmax"], info["ncpu"]).T

        # Read boundary grids if any
        self.offsets["i"] += 10 * info["levelmax"]
        self.offsets["n"] += 3
        if self.meta["nboundary"] > 0:
            self.offsets["i"] += (2 * self.meta["nboundary"] * info["levelmax"])
            # self.offsets["n"] += 4
            self.meta["ngridlevel"][info["ncpu"]:info["ncpu"] +
                                    self.meta["nboundary"], :] = np.array(
                                        utils.read_binary_data(
                                            fmt="{}i".format(self.meta["nboundary"] *
                                                             info["levelmax"]),
                                            content=self.bytes,
                                            offsets=self.offsets)).reshape(
                                                info["levelmax"],
                                                self.meta["nboundary"]).T
            self.offsets["n"] += 2

        # Determine bound key precision
        self.offsets["i"] += 5
        self.offsets["s"] += 128
        [key_size] = utils.read_binary_data(fmt="i",
                                            content=self.bytes,
                                            offsets=self.offsets,
                                            skip_head=False,
                                            increment=False)

        # Offset for AMR
        self.offsets["i"] += 3 * ncoarse
        self.offsets["n"] += 3
        self.offsets["s"] += key_size

    def read_level_header(self, ilevel, twotondim):
        # Geometry
        self.dxcell = 0.5**(ilevel + 1)
        for ind in range(twotondim):
            iz = int((ind) / 4)
            iy = int((ind - 4 * iz) / 2)
            ix = int((ind - 2 * iy - 4 * iz))
            self.xcent[ind, 0] = (float(ix) - 0.5) * self.dxcell
            self.xcent[ind, 1] = (float(iy) - 0.5) * self.dxcell
            self.xcent[ind, 2] = (float(iz) - 0.5) * self.dxcell

    def read_cacheline_header(self, ncache, ndim):
        # xg: grid coordinates
        self.offsets['i'] += ncache * 3
        self.offsets['n'] += 3
        for n in range(ndim):
            self.xg[:ncache, n] = utils.read_binary_data(fmt="{}d".format(ncache),
                                                         content=self.bytes,
                                                         offsets=self.offsets)

        # son indices
        self.offsets['i'] += ncache * (1 + 2 * ndim)
        self.offsets['n'] += 1 + 2 * ndim

    def read_variables(self, ncache, ind, ilevel, cpuid, info):

        self.son[:ncache, ind] = utils.read_binary_data(fmt="{}i".format(ncache),
                                                        content=self.bytes,
                                                        offsets=self.offsets)

        if self.variables["level"]["read"]:
            self.variables["level"]["buffer"]._array[:ncache, ind] = ilevel + 1
        for n in range(info["ndim"]):
            key = "xyz_" + "xyz"[n]
            if self.variables[key]["read"]:
                self.variables[key]["buffer"]._array[:ncache, ind] = (
                    self.xg[:ncache, n] + self.xcent[ind, n] - self.meta["xbound"][n]
                ) * info["boxlen"] * self.variables[key]["unit"].magnitude
        if self.variables["dx"]["read"]:
            self.variables["dx"]["buffer"]._array[:ncache, ind] = self.dxcell * info[
                "boxlen"] * self.variables["dx"]["unit"].magnitude
        if self.variables["cpu"]["read"]:
            self.variables["cpu"]["buffer"]._array[:ncache, ind] = cpuid + 1

        # Note: use lmax here instead of levelmax because the user might not
        # want to load all levels. levelmax is always the max level in the
        # entire simulation.
        self.ref[:ncache, ind] = np.logical_not(
            np.logical_and(self.son[:ncache, ind] > 0, ilevel < info["lmax"] - 1))

    def make_conditions(self, select, ncache):
        conditions = super().make_conditions(select, ncache)
        conditions.update({"leaf": self.ref[:ncache, :]})
        return conditions

    def read_footer(self, ncache, twotondim):
        # Increment offsets with remainder of the file
        self.offsets['i'] += ncache * 2 * twotondim
        self.offsets['n'] += 2 * twotondim

    def step_over(self, ncache, twotondim, ndim):
        self.offsets['i'] += ncache * (4 + 3 * twotondim + 2 * ndim)
        self.offsets['d'] += ncache * ndim
        self.offsets['n'] += 4 + 3 * twotondim + 3 * ndim



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
from typing import Union
from ..core import Plot, Array
from .render import render


def plot(x: Array,
         *y: Union[Array, dict],
         logx: bool = False,
         logy: bool = False,
         loglog: bool = False,
         filename: str = None,
         title: str = None,
         xmin: float = None,
         xmax: float = None,
         ymin: float = None,
         ymax: float = None,
         legend=True,
         ax: object = None,
         **kwargs) -> Plot:
    """
    Make a 1D plot with two variables as input.

    This function has an API very close to that of matplotlib's ``scatter`` function.
    For the documentation of any parameters that are not listed below, see
    https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.scatter.html.

    :param x: Array to use for x positions along the horizontal dimension.

    :param y: Array(s) to use for y positions along the vertical dimension.

    :param logx: If ``True``, use logarithmic scaling on the horizontal axis.
        Default is ``False``.

    :param logy: If ``True``, use logarithmic scaling on the vertical axis.
        Default is ``False``.

    :param loglog: If ``True``, use logarithmic scaling on the horizontal and
        vertical axes. Default is ``False``.

    :param filename: If specified, the returned figure is also saved to file.
        Default is ``None``.

    :param title: The title of the figure. Default is ``None``.

    :param xmin: Minimum value for the horizontal axis. Default is ``None``.

    :param xmax: Maximum value for the horizontal axis. Default is ``None``.

    :param ymin: Minimum value for the vertical axis. Default is ``None``.

    :param ymax: Maximum value for the vertical axis. Default is ``None``.

    :param ax: A matplotlib axes inside which the figure will be plotted.
        Default is ``None``, in which case some new axes a created.
    """
    if loglog:
        logx = logy = True

    to_plot = []
    yaxis_unit = None
    xvals = None

    if isinstance(x, dict):
        to_plot.append({
            "x": x["x"].norm,
            "y": x["y"].norm,
            "params": {
                **kwargs
            },
            "unit": x["y"].unit.units,
            "name": x["y"].name
        })
        yaxis_unit = x["y"].unit
    else:
        xvals = x.norm

    if isinstance(y, Array):
        y = [y]

    for layer in y:

        if xvals is not None:
            layer_x = xvals
        else:
            layer_x = to_plot[0]["x"]

        if isinstance(layer, dict):
            layer_dict = {
                "x": layer["x"].norm,
                "y": layer["y"].norm,
                "params": {
                    **kwargs
                },
                "unit": layer["y"].unit.units,
                "name": layer["y"].name
            }
            layer_dict["x"] = layer["x"].norm if "x" in layer else layer_x
            to_plot.append(layer_dict)
            layer_unit = layer["y"].unit
        else:
            to_plot.append({
                "x": layer_x,
                "y": layer.norm,
                "params": {
                    **kwargs
                },
                "unit": layer.unit.units,
                "name": layer.name
            })
            layer_unit = layer.unit

        if yaxis_unit is None:
            yaxis_unit = layer_unit
        else:
            if layer_unit != yaxis_unit:
                raise RuntimeError(
                    "Different layers in 1D plots must all have the same unit.")

    figure = render(logx=logx, logy=logy, ax=ax)
    for item in to_plot:
        sorting = np.argsort(item["x"].values)
        figure["ax"].plot(item["x"].values[sorting],
                          item['y'].values[sorting],
                          label=item["name"],
                          **item["params"])

    figure["ax"].set_xlabel(to_plot[0]['x'].label)
    figure["ax"].set_ylabel(to_plot[0]['y'].label)
    figure["ax"].set_xlim(xmin, xmax)
    figure["ax"].set_ylim(ymin, ymax)
    figure["ax"].set_title(title)
    if legend and len(to_plot) > 1:
        figure["ax"].legend()

    return Plot(layers=to_plot, fig=figure["fig"], ax=figure["ax"], filename=filename)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

# Import the config from "/home/user/.osyris/conf.py if it exists.
# If it doesn't, try to create one by copying the default from the source.
# If that fails, just load the default.
import os
import sys
from shutil import copyfile
from . import defaults as default_config

user_config_dir = os.path.join(os.path.expanduser("~"), ".osyris")
this_dir = os.path.dirname(os.path.abspath(__file__))
if not os.path.exists(user_config_dir):
    os.mkdir(user_config_dir)
user_config_file = os.path.join(user_config_dir, "config_osyris.py")
if not os.path.exists(user_config_file):
    copyfile(os.path.join(this_dir, "defaults.py"), user_config_file)
sys.path.append(user_config_dir)

import config_osyris as user_config  # noqa: E402


# Create a dummy class
class Config:
    pass


if hasattr(user_config, 'ureg'):
    units = getattr(user_config, 'ureg')
else:
    units = getattr(default_config, 'ureg')

config = Config()

# Import list of object from user config if present, if not, load from defaults.
objects = ['parameters', 'get_unit', 'additional_units', 'additional_variables']
for obj in objects:
    if hasattr(user_config, obj):
        setattr(config, obj, getattr(user_config, obj))
    else:
        setattr(config, obj, getattr(default_config, obj))

config.additional_units()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

from pint.quantity import Quantity
from ..core import Plot, Array
from .. import units
from .render import render
from .parser import parse_layer


def scatter(x: Array,
            y: Array,
            logx: bool = False,
            logy: bool = False,
            loglog: bool = False,
            norm: str = None,
            filename: str = None,
            title: str = None,
            xmin: float = None,
            xmax: float = None,
            ymin: float = None,
            ymax: float = None,
            vmin: float = None,
            vmax: float = None,
            ax: object = None,
            **kwargs) -> Plot:
    """
    Make a 2D scatter plot with two variables as input.

    This function has an API very close to that of matplotlib's ``scatter`` function.
    For the documentation of any parameters that are not listed below, see
    https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.scatter.html.

    :param x: Array to use for scatter point positions along the horizontal dimension.

    :param y: Array to use for scatter point positions along the vertical dimension.

    :param logx: If ``True``, use logarithmic scaling on the horizontal axis.
        Default is ``False``.

    :param logy: If ``True``, use logarithmic scaling on the vertical axis.
        Default is ``False``.

    :param loglog: If ``True``, use logarithmic scaling on the horizontal and
        vertical axes. Default is ``False``.

    :param norm: The colormap normalization. Possible values are ``'linear'`` and
        ``'log'``. Default is ``None`` (= ``'linear'``).

    :param filename: If specified, the returned figure is also saved to file.
        Default is ``None``.

    :param title: The title of the figure. Default is ``None``.

    :param xmin: Minimum value for the horizontal axis. Default is ``None``.

    :param xmax: Maximum value for the horizontal axis. Default is ``None``.

    :param ymin: Minimum value for the vertical axis. Default is ``None``.

    :param ymax: Maximum value for the vertical axis. Default is ``None``.

    :param vmin: Minimum value for colorbar range. Default is ``None``.

    :param vmax: Maximum value for colorbar range. Default is ``None``.

    :param ax: A matplotlib axes inside which the figure will be plotted.
        Default is ``None``, in which case some new axes a created.
    """
    if loglog:
        logx = logy = True

    xvals = x.norm.values
    yvals = y.norm.values

    _, _, params = parse_layer(layer=None, norm=norm, vmin=vmin, vmax=vmax, **kwargs)
    to_render = [{
        "data": None,
        "mode": "scatter",
        "params": params,
    }]
    if "c" in params and not isinstance(params["c"], str):
        to_render[0].update({"unit": params["c"].unit.units, "name": params["c"].name})
        params["c"] = params["c"].norm.values
    if "s" in params:
        unit = None
        if isinstance(params["s"], Array):
            unit = params["s"].unit.units
        if isinstance(params["s"], Quantity):
            unit = params["s"].units
        if unit is not None:
            if unit != units.dimensionless:
                if x.unit.units != y.unit.units:
                    raise RuntimeError("Scatter: an Array with a unit was supplied "
                                       "as a size, but the units of the x and y "
                                       "Arrays do not agree. The size must either "
                                       "be a float or a dimensionless Array.")
                params["s"] = params["s"].to(x.unit.units)

    figure = render(x=xvals, y=yvals, data=to_render, logx=logx, logy=logy, ax=ax)

    figure["ax"].set_xlabel(x.label)
    figure["ax"].set_ylabel(y.label)
    figure["ax"].set_xlim(xmin, xmax)
    figure["ax"].set_ylim(ymin, ymax)

    return Plot(x=xvals,
                y=yvals,
                layers=to_render,
                fig=figure["fig"],
                ax=figure["ax"],
                filename=filename)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
import os
from ..core import Array, Datagroup
from .reader import ReaderKind
from .. import units
from . import utils


class SinkReader:
    def __init__(self):
        self.kind = ReaderKind.SINK
        self.initialized = False

    def initialize(self, meta, select):
        if select is False:
            return
        sink = Datagroup()
        sink_file = utils.generate_fname(meta["nout"],
                                         meta["path"],
                                         ftype="sink",
                                         cpuid=0,
                                         ext=".csv")
        if not os.path.exists(sink_file):
            return

        if os.path.getsize(sink_file) == 0:
            # This is an empty sink file
            return sink
        else:
            sink_data = np.atleast_2d(np.loadtxt(sink_file, delimiter=',', skiprows=2))

        with open(sink_file, 'r') as f:
            key_list = f.readline()
            unit_combinations = f.readline()

        key_list = key_list.lstrip(' #').rstrip('\n').split(',')
        unit_combinations = unit_combinations.lstrip(' #').rstrip('\n').split(',')

        # Parse units
        unit_list = []
        for u in unit_combinations:
            m = meta['unit_d'] * meta['unit_l']**3 * units.g  # noqa: F841
            l = meta['unit_l'] * units.cm  # noqa: F841, E741
            t = meta['unit_t'] * units.s  # noqa: F841
            if u.strip() == '1':
                unit_list.append(1.0 * units.dimensionless)
            else:
                unit_list.append(eval(u.replace(' ', '*')))

        sink = Datagroup()
        for i, (key, unit) in enumerate(zip(key_list, unit_list)):
            sink[key] = Array(values=sink_data[:, i] * unit.magnitude, unit=unit.units)
            if unit_combinations[i] == 'l':
                sink[key] = sink[key].to(meta["scale"])
        utils.make_vector_arrays(sink, ndim=meta["ndim"])
        return sink



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

import matplotlib.pyplot as plt

from . import wrappers
from .. import config
from ..core.tools import make_label


def render(x=None, y=None, data=None, logx=False, logy=False, ax=None):
    """
    Use matplotlib to plot histogram, slice or column density maps
    """

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    if logx:
        ax.set_xscale("log")
    if logy:
        ax.set_yscale("log")

    out = {"fig": fig, "ax": ax}
    if data is None:
        return out

    function_map = {
        "vec": "quiver",
        "vector": "quiver",
        "stream": "streamplot",
        "lic": "line_integral_convolution",
        None: config.parameters["render_mode"],
        "image": "pcolormesh",
        "imshow": "pcolormesh"
    }

    mpl_objects = []
    for item in data:
        func = item["mode"]
        if func in function_map:
            func = function_map[func]

        if "cbar" in item["params"]:
            cbar = item["params"]["cbar"]
            del item["params"]["cbar"]
        else:
            cbar = True

        if func == "line_integral_convolution" and "color" in item["params"]:
            cblabel = make_label(name=item["params"]["color"].name,
                                 unit=item["params"]["color"].unit.units)
        else:
            cblabel = make_label(name=item.get("name", ""), unit=item.get("unit", ""))

        mpl_objects.append(
            getattr(wrappers, func)(ax=ax,
                                    x=x,
                                    y=y,
                                    z=item["data"],
                                    cbar=cbar,
                                    cblabel=cblabel,
                                    **item["params"]))

    out["objects"] = mpl_objects
    return out



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)
import osyris
import pytest


def test_datagroup_creation():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = osyris.Datagroup()
    dg['a'] = a
    dg['b'] = b
    assert 'a' in dg.keys()
    assert 'b' in dg.keys()
    assert dg['a'].name == 'a'
    assert dg['b'].name == 'b'
    assert len(dg) == 2


def test_datagroup_creation_from_dict():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = osyris.Datagroup({'a': a, 'b': b})
    assert 'a' in dg.keys()
    assert 'b' in dg.keys()
    assert dg['a'].name == 'a'
    assert dg['b'].name == 'b'
    assert len(dg) == 2


def test_datagroup_insert_vector():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[[6., 7., 8.], [9., 10., 11.], [12., 13., 14.],
                             [15., 16., 17.], [18., 19., 20.]],
                     unit='s')
    dg = osyris.Datagroup()
    dg['a'] = a
    dg['b'] = b
    assert 'a' in dg.keys()
    assert 'b' in dg.keys()
    assert dg['a'].name == 'a'
    assert dg['b'].name == 'b'
    assert len(dg) == 2


def test_datagroup_bad_length_insertion():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9.], unit='s')
    dg = osyris.Datagroup()
    dg['a'] = a
    with pytest.raises(RuntimeError):
        dg['b'] = b


def test_datagroup_delitem():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = osyris.Datagroup()
    dg['a'] = a
    dg['b'] = b
    del dg['a']
    assert 'a' not in dg.keys()
    assert 'b' in dg.keys()
    assert len(dg) == 1


def test_datagroup_slice():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = osyris.Datagroup()
    dg['a'] = a
    dg['b'] = b
    expected = osyris.Datagroup({
        'a': osyris.Array(values=[2.], unit='m'),
        'b': osyris.Array(values=[7.], unit='s')
    })
    sliced = dg[1]
    assert sliced['a'] == expected['a']
    assert sliced['b'] == expected['b']


def test_datagroup_slice_range():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg = osyris.Datagroup()
    dg['a'] = a
    dg['b'] = b
    expected = osyris.Datagroup({
        'a': osyris.Array(values=[2., 3., 4.], unit='m'),
        'b': osyris.Array(values=[7., 8., 9.], unit='s')
    })
    sliced = dg[1:4]
    assert all(sliced['a'] == expected['a'])
    assert all(sliced['b'] == expected['b'])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)
import os
from .reader import Reader, ReaderKind
from . import utils


class GravReader(Reader):
    def __init__(self):
        super().__init__(kind=ReaderKind.AMR)

    def initialize(self, meta, select):
        self.initialized = False
        if select is False:
            return

        fname = utils.generate_fname(meta["nout"], meta["path"], ftype="grav", cpuid=1)
        # Check if self-gravity files exist
        if not os.path.exists(fname):
            return
        # Add gravity fields
        descriptor = {"potential": "d"}
        for n in range(meta["ndim"]):
            descriptor["acceleration_" + "xyz"[n]] = "d"

        self.descriptor_to_variables(descriptor=descriptor, meta=meta, select=select)
        self.initialized = True

    def read_header(self, info):
        self.offsets["i"] += 4
        self.offsets["n"] += 4

    def read_domain_header(self):
        self.offsets['n'] += 2
        self.offsets['i'] += 2



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
import numpy.ma as ma
from pint.quantity import Quantity
from typing import Union
from .slice import get_slice_direction
from .render import render
from .scatter import scatter
from .parser import parse_layer
from ..core import Plot, Array
from ..core.tools import apply_mask
from .utils import evaluate_on_grid


def _add_scatter(to_scatter, origin, dir_vecs, dx, dy, ax):
    xyz = to_scatter[0]["data"] - origin
    viewport = max(dx.magnitude, dy.magnitude)
    radius = None
    if "s" in to_scatter[0]["params"]:
        size = to_scatter[0]["params"]["s"]
        if isinstance(size, Array) or isinstance(size, Quantity):
            radius = size.to(dx.units)
            to_scatter[0]["params"]["s"] = radius
    if radius is None:
        # Fudge factor to select sinks close to the plane
        radius = Array(values=viewport * 0.05, unit=dx.units)
    dist1 = np.sum(xyz * dir_vecs[0], axis=1)
    global_selection = np.arange(len(to_scatter[0]["data"]))
    select = np.ravel(np.where(np.abs(dist1) <= radius))
    global_selection = global_selection[select]
    if len(select) > 0:
        # Project coordinates onto the plane by taking dot product with axes vectors
        coords = xyz[select]
        datax = np.inner(coords, dir_vecs[1])
        datay = np.inner(coords, dir_vecs[2])
        if dx is not None:
            # Limit selection further by using distance from center
            dist2 = coords
            select2 = np.ravel(
                np.where(np.abs(dist2.norm.values) <= viewport * 0.6 * np.sqrt(2.0)))
            datax = datax[select2]
            datay = datay[select2]
            global_selection = global_selection[select2]
        if "c" in to_scatter[0]["params"]:
            # TODO: also check that parents are the same to ensure size match?
            if isinstance(to_scatter[0]["params"]["c"], Array):
                to_scatter[0]["params"]["c"] = to_scatter[0]["params"]["c"][
                    global_selection]
        scatter(x=datax, y=datay, ax=ax, **to_scatter[0]["params"])


def map(*layers,
        direction: Union[str, list] = "z",
        dx: Quantity = None,
        dy: Quantity = None,
        dz: Quantity = None,
        filename: str = None,
        title: str = None,
        plot: bool = True,
        mode: str = None,
        norm: str = None,
        vmin: float = None,
        vmax: float = None,
        origin: Array = None,
        resolution: Union[int, dict] = None,
        operation: str = "sum",
        ax: object = None,
        **kwargs) -> Plot:
    """
    Create a 2D spatial map of a region inside a simulation domain.
    By default, the map represents a plane with zero thickness.
    A thick slab or cube can also be computed by specifying a thickness via the
    ``dz`` argument. In this case, the resulting 3D box is integrated along the ``z``
    direction before being sent to the image rendering.

    :param layers: Dicts or Arrays representing the quantities to be mapped onto the
        generated image.

    :param direction: The vector normal to the map. Possible choices are:

       * ``'x'``, ``'y'``, or ``'z'`` representing the cartesian axes
       * a list of 3 numbers representing the components of the vector,
         e.g. ``[1, 0.5, 2]``
       * ``'top'`` or ``'side'`` for automatic top or side view of a disk, according to
         the angular momentum computed around the center of the plotted region

    :param dx: The horizontal size of the plotted region. Default is ``None``,
        in which case the entire horizontal range of the simulation domain is plotted.

    :param dy: The vertical size of the plotted region. If not specified, it will
        either be equal to ``dx`` if ``dx`` is not ``None``, or the entire vertical
        range of the simulation domain if ``dx`` is ``None``. Default is ``None``.

    :param dz: The depth range over which the ``z`` dimension is to be integrated.
        Default is ``None``, in which case a plane with no thickness is plotted.

    :param filename: If specified, the returned figure is also saved to file.
        Default is ``None``.

    :param title: The title of the figure. Default is ``None``.

    :param plot: Make a plot if ``True``. If not, just return the ``Plot`` object
        containing the data that would be used to generate the plot.
        Default is ``True``.

    :param mode: The rendering mode for the map. Possible choices are ``'image'``,
        ``'contourf'``, and ``'contour'`` for scalar Arrays, ``'vec'`` and
        ``'stream'`` for vector quantities. Default is ``None``, which selects the
        ``render_mode`` set in the user configuration file (``'image'`` by default).

    :param norm: The colormap normalization. Possible values are ``'linear'`` and
        ``'log'``. Default is ``None`` (= ``'linear'``).

    :param vmin: Minimum value for colorbar range. Default is ``None``.

    :param vmax: Maximum value for colorbar range. Default is ``None``.

    :param origin: An Array describing the position of the center of the map
        (with 2 or 3 components depending on the dimensionality of the simulation).

    :param resolution: Resolution of the generated map. This can either be an
        integer or a dict. In the case of an integer, it represents the number of
        pixels used for the horizontal and vertical dimensions. For a dictionary,
        the following syntax should be used: ``resolution={'x': 128, 'y': 192}``.
        Default is ``256``.

    :param operation: The operation to apply along the ``z`` dimension if ``dz`` is
        not ``None``. Possible values are ``'sum'``, ``'mean'``, ``'min'``, and
        ``'max'``. Default is ``'sum'``.

    :param ax: A matplotlib axes inside which the figure will be plotted.
        Default is ``None``, in which case some new axes a created.
    """

    if isinstance(layers, Array):
        layers = [layers]

    to_process = []
    to_render = []
    to_scatter = []
    for layer in layers:
        data, settings, params = parse_layer(layer=layer,
                                             mode=mode,
                                             norm=norm,
                                             vmin=vmin,
                                             vmax=vmax,
                                             **kwargs)
        if settings["mode"] == "scatter":
            to_scatter.append({"data": data, "params": params})
        else:
            to_process.append(data)
            to_render.append({
                "mode": settings["mode"],
                "params": params,
                "unit": data.unit.units,
                "name": data.name
            })

    dataset = to_process[0].parent.parent

    thick = dz is not None

    # Set window size
    if dy is None:
        dy = dx
    if dz is None:
        dz = dx
    if dx is not None and not isinstance(dx, Quantity):
        dx *= dataset["amr"]["xyz"].unit
    if dy is not None and not isinstance(dy, Quantity):
        dy *= dataset["amr"]["xyz"].unit
    if dz is not None and not isinstance(dz, Quantity):
        dz *= dataset["amr"]["xyz"].unit

    dir_vecs, origin = get_slice_direction(direction=direction,
                                           dataset=dataset,
                                           dx=dx,
                                           dy=dy,
                                           origin=origin)

    # Distance to the plane
    diagonal = np.sqrt(dataset.meta["ndim"])
    xyz = dataset["amr"]["xyz"] - origin
    selection_distance = 0.5 * diagonal * (dz if thick else dataset["amr"]["dx"])
    dist_to_plane = np.sum(xyz * dir_vecs[0], axis=1)
    # Create an array of indices to allow further narrowing of the selection below
    global_indices = np.arange(len(dataset["amr"]["dx"]))
    # Select cells close to the plane, including factor of sqrt(ndim)
    close_to_plane = np.ravel(np.where(np.abs(dist_to_plane) <= selection_distance))
    indices_close_to_plane = global_indices[close_to_plane]

    if len(indices_close_to_plane) == 0:
        raise RuntimeError("No cells were selected to construct the column density. "
                           "The resulting figure would be empty.")

    xmin = None
    if dx is not None:
        xmin = -0.5 * dx.magnitude
        xmax = xmin + dx.magnitude
        ymin = -0.5 * dy.magnitude
        ymax = ymin + dy.magnitude
        zmin = -0.5 * dz.magnitude
        zmax = zmin + dz.magnitude
        # Limit selection further by using distance from center
        radial_distance = xyz[indices_close_to_plane] - 0.5 * dataset["amr"]["dx"][
            indices_close_to_plane] * diagonal
        radial_selection = np.ravel(
            np.where(
                np.abs(radial_distance.norm.values) <=
                max(dx.magnitude, dy.magnitude, dz.magnitude) * 0.6 * diagonal))
        indices_close_to_plane = indices_close_to_plane[radial_selection]

    # Project coordinates onto the plane by taking dot product with axes vectors
    coords = xyz[indices_close_to_plane]
    datax = np.inner(coords, dir_vecs[1])
    datay = np.inner(coords, dir_vecs[2])
    dataz = np.inner(coords, dir_vecs[0])
    datadx = dataset["amr"]["dx"][indices_close_to_plane] * 0.5

    if xmin is None:
        xmin = (datax - datadx).min().values
        xmax = (datax + datadx).max().values
        ymin = (datay - datadx).min().values
        ymax = (datay + datadx).max().values
        zmin = (dataz - datadx).min().values
        zmax = (dataz + datadx).max().values
        dx = (xmax - xmin) * datadx.unit.units
        dy = (ymax - ymin) * datadx.unit.units

    scalar_layer = []
    to_binning = []  # contains the variables in cells close to the plane
    for ind in range(len(to_process)):
        if to_render[ind]["mode"] in ["vec", "stream", "lic"]:
            if to_process[ind].ndim < 3:
                uv = to_process[ind].array[indices_close_to_plane]
            else:
                uv = np.inner(
                    to_process[ind].array.take(indices_close_to_plane, axis=0),
                    dir_vecs[1:])
            w = None
            if "color" in to_render[ind]["params"]:
                if isinstance(to_render[ind]["params"]["color"], Array):
                    w = to_render[ind]["params"]["color"].norm.values
                elif isinstance(to_render[ind]["params"]["color"], np.ndarray):
                    w = to_render[ind]["params"]["color"]
            if w is None:
                w = np.linalg.norm(uv, axis=1)
            else:
                w = w.take(indices_close_to_plane, axis=0)
            to_binning.append(apply_mask(uv[:, 0]))
            to_binning.append(apply_mask(uv[:, 1]))
            to_binning.append(w)
            scalar_layer.append(False)
        else:
            to_binning.append(
                apply_mask(to_process[ind].norm.values[indices_close_to_plane]))
            scalar_layer.append(True)

    # Create a grid of pixel centers
    default_resolution = 256
    if resolution is None:
        resolution = default_resolution
    if isinstance(resolution, int):
        resolution = {'x': resolution, 'y': resolution}
    else:
        for xy in 'xy':
            if xy not in resolution:
                resolution[xy] = default_resolution
    xspacing = (xmax - xmin) / resolution['x']
    yspacing = (ymax - ymin) / resolution['y']

    if 'z' not in resolution:
        if not thick:
            resolution['z'] = 1
        else:
            # Try to keep spacing similar to other spacings
            resolution['z'] = round((zmax - zmin) / (0.5 * (xspacing + yspacing)))
    zspacing = (zmax - zmin) / resolution['z']

    xcenters = np.linspace(xmin + 0.5 * xspacing, xmax - 0.5 * xspacing,
                           resolution['x'])
    ycenters = np.linspace(ymin + 0.5 * yspacing, ymax - 0.5 * yspacing,
                           resolution['y'])
    zcenters = np.linspace(zmin + 0.5 * zspacing, zmax - 0.5 * zspacing,
                           resolution['z'])

    xg, yg, zg = np.meshgrid(xcenters, ycenters, zcenters, indexing='ij')
    xgrid = xg.T
    ygrid = yg.T
    zgrid = zg.T
    pixel_positions = xgrid.reshape(xgrid.shape + (1, )) * dir_vecs[1] + ygrid.reshape(
        ygrid.shape + (1, )) * dir_vecs[2] + zgrid.reshape(zgrid.shape +
                                                           (1, )) * dir_vecs[0]

    # Evaluate the values of the data layers at the grid positions
    binned = evaluate_on_grid(cell_positions_in_new_basis=np.array(
        [apply_mask(datax.array),
         apply_mask(datay.array),
         apply_mask(dataz.array)]).T,
                              cell_positions_in_original_basis=coords.array,
                              cell_values=np.array(to_binning),
                              cell_sizes=datadx.array,
                              grid_lower_edge_in_new_basis=np.array([xmin, ymin, zmin]),
                              grid_spacing_in_new_basis=np.array(
                                  [xspacing, yspacing, zspacing]),
                              grid_positions_in_original_basis=pixel_positions,
                              ndim=dataset.meta["ndim"])

    # Apply operation along depth
    binned = getattr(binned, operation)(axis=1)

    # Handle thick maps
    if thick:
        binned *= zspacing
        for layer in to_render:
            layer["unit"] = (Array(values=1, unit=layer["unit"]) *
                             dataz.unit).unit.units

    # Mask NaN values
    mask = np.isnan(binned[-1, ...])
    mask_vec = np.broadcast_to(mask.reshape(*mask.shape, 1), mask.shape + (3, ))

    # Now we fill the arrays to be sent to the renderer, also constructing vectors
    counter = 0
    for ind in range(len(to_render)):
        if scalar_layer[ind]:
            to_render[ind]["data"] = ma.masked_where(mask,
                                                     binned[counter, ...],
                                                     copy=False)
            counter += 1
        else:
            to_render[ind]["data"] = ma.masked_where(mask_vec,
                                                     np.array([
                                                         binned[counter, ...].T,
                                                         binned[counter + 1, ...].T,
                                                         binned[counter + 2, ...].T
                                                     ]).T,
                                                     copy=False)
            counter += 3

    to_return = {
        "x": xcenters,
        "y": ycenters,
        "layers": to_render,
        "filename": filename
    }
    if plot:
        # Render the map
        figure = render(x=xcenters, y=ycenters, data=to_render, ax=ax)
        figure["ax"].set_xlabel(dataset["amr"]["xyz"].x.label)
        figure["ax"].set_ylabel(dataset["amr"]["xyz"].y.label)
        if ax is None:
            figure["ax"].set_aspect("equal")

        # Add scatter layer
        if len(to_scatter) > 0:
            _add_scatter(to_scatter=to_scatter,
                         origin=origin,
                         dir_vecs=dir_vecs,
                         dx=dx,
                         dy=dy,
                         ax=figure["ax"])

        figure["ax"].set_xlim(xmin, xmax)
        figure["ax"].set_ylim(ymin, ymax)

        to_return.update({"fig": figure["fig"], "ax": figure["ax"]})

    return Plot(**to_return)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)


class Plot:
    def __init__(self, x=None, y=None, layers=None, fig=None, ax=None, filename=None):
        self.x = x
        self.y = y
        self.layers = layers
        self.fig = fig
        self.ax = ax
        if filename is not None:
            self.fig.savefig(filename, bbox_inches='tight')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)
import osyris
import numpy as np
import pint
import pytest


def test_1d_array():
    a = np.arange(100.)
    array = osyris.Array(values=a, unit='m')
    assert array.unit == osyris.units('m')
    assert len(array) == len(a)
    assert array.shape == a.shape
    assert np.allclose(array.values, a)


def test_2d_array():
    a = np.random.random([100, 3])
    array = osyris.Array(values=a, unit='m')
    assert array.unit == osyris.units('m')
    assert np.allclose(array.x.values, a[:, 0])
    assert np.allclose(array.y.values, a[:, 1])
    assert np.allclose(array.z.values, a[:, 2])


def test_equal():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    c = osyris.Array(values=[100., 200., 300., 400., 500.], unit='cm')
    assert all(a == b)
    assert all(a == c)


def test_not_equal():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[1., 2., 3., 4., 5.], unit='cm')
    c = osyris.Array(values=[100., 200., 300., 400., 500.], unit='m')
    assert all(a != b)
    assert all(a != c)


def test_addition():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='m')
    expected = osyris.Array(values=[7., 9., 11., 13., 15.], unit='m')
    assert all(a + b == expected)


def test_addition_bad_units():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='s')
    with pytest.raises(pint.errors.DimensionalityError):
        _ = a + b
    with pytest.raises(TypeError):
        _ = a + 3.0


def test_addition_quantity():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = 3.5 * osyris.units('m')
    expected = osyris.Array(values=[4.5, 5.5, 6.5, 7.5, 8.5], unit='m')
    assert all(a + b == expected)


def test_addition_inplace():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='m')
    expected = osyris.Array(values=[7., 9., 11., 13., 15.], unit='m')
    a += b
    assert all(a == expected)


def test_addition_quantity_inplace():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = 3.5 * osyris.units('m')
    expected = osyris.Array(values=[4.5, 5.5, 6.5, 7.5, 8.5], unit='m')
    a += b
    assert all(a == expected)


def test_subtraction():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='m')
    expected = osyris.Array(values=[5., 5., 5., 5., 5.], unit='m')
    assert all(b - a == expected)


def test_subtraction_bad_units():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='s')
    with pytest.raises(pint.errors.DimensionalityError):
        _ = a - b
    with pytest.raises(TypeError):
        _ = a - 3.0


def test_subtraction_quantity():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = 3.5 * osyris.units('m')
    expected = osyris.Array(values=[-2.5, -1.5, -0.5, 0.5, 1.5], unit='m')
    assert all(a - b == expected)


def test_subtraction_inplace():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='m')
    expected = osyris.Array(values=[5., 5., 5., 5., 5.], unit='m')
    b -= a
    assert all(b == expected)


def test_subtraction_quantity_inplace():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = 3.5 * osyris.units('m')
    expected = osyris.Array(values=[-2.5, -1.5, -0.5, 0.5, 1.5], unit='m')
    a -= b
    assert all(a == expected)


def test_multiplication():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='m')
    expected = osyris.Array(values=[6., 14., 24., 36., 50.], unit='m*m')
    assert all(a * b == expected)


def test_multiplication_float():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = 3.0
    expected = osyris.Array(values=[3., 6., 9., 12., 15.], unit='m')
    assert all(a * b == expected)


def test_multiplication_quantity():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = 3.5 * osyris.units('s')
    expected = osyris.Array(values=[3.5, 7.0, 10.5, 14.0, 17.5], unit='m*s')
    assert all(a * b == expected)


def test_multiplication_inplace():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='m')
    expected = osyris.Array(values=[6., 14., 24., 36., 50.], unit='m*m')
    a *= b
    assert all(a == expected)


def test_multiplication_float_inplace():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = 3.0
    expected = osyris.Array(values=[3., 6., 9., 12., 15.], unit='m')
    a *= b
    assert all(a == expected)


def test_multiplication_quantity_inplace():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = 3.5 * osyris.units('s')
    expected = osyris.Array(values=[3.5, 7.0, 10.5, 14.0, 17.5], unit='m*s')
    a *= b
    assert all(a == expected)


def test_division():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='s')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='m')
    expected = osyris.Array(values=[600., 350., 800. / 3., 225.0, 200.], unit='cm/s')
    assert all(b / a == expected)


def test_division_float():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='s')
    b = 3.0
    expected = osyris.Array(values=[1. / 3., 2. / 3., 1., 4. / 3., 5. / 3.], unit='s')
    assert all(a / b == expected)


def test_division_quantity():
    a = osyris.Array(values=[0., 2., 4., 6., 200.], unit='s')
    b = 2.0 * osyris.units('s')
    expected = osyris.Array(values=[0., 1., 2., 3., 100.], unit='dimensionless')
    assert all(a / b == expected)


def test_division_inplace():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='s')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='m')
    expected = osyris.Array(values=[600., 350., 800. / 3., 225.0, 200.], unit='cm/s')
    b /= a
    assert all(b == expected)


def test_division_float_inplace():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='s')
    b = 3.0
    expected = osyris.Array(values=[1. / 3., 2. / 3., 1., 4. / 3., 5. / 3.], unit='s')
    a /= b
    assert all(a == expected)


def test_division_quantity_inplace():
    a = osyris.Array(values=[0., 2., 4., 6., 200.], unit='s')
    b = 2.0 * osyris.units('s')
    expected = osyris.Array(values=[0., 1., 2., 3., 100.], unit='dimensionless')
    a /= b
    assert all(a == expected)


def test_norm():
    a2d = osyris.Array(values=np.array([[1., 2.], [3., 4.], [5., 6.], [7., 8.]]),
                       unit='s')
    a3d = osyris.Array(values=np.array([[1., 2., 3.], [4., 5., 6.]]), unit='g')
    assert all(a2d.norm == osyris.Array(values=np.sqrt([5., 25., 61., 113.]), unit='s'))
    assert all(a3d.norm == osyris.Array(values=np.sqrt([14., 77.]), unit='g'))


def test_broadcast():
    a1d = osyris.Array(values=np.array([1., 2., 3., 4., 5.]), unit='s')
    a3d = osyris.Array(values=np.array([[1., 2., 3.], [4., 5., 6.], [7., 8., 9.],
                                        [10., 11., 12.], [13., 14., 15.]]),
                       unit='g')
    expected = osyris.Array(values=np.array([[1., 2., 3.], [8., 10., 12.],
                                             [21., 24., 27.], [40., 44., 48.],
                                             [65., 70., 75.]]),
                            unit='g*s')
    assert all(np.ravel(a1d * a3d == expected))


def test_power():
    a = osyris.Array(values=[1., 2., 4., 6., 200.], unit='s')
    expected = osyris.Array(values=[1., 8., 64., 216., 8.0e6], unit='s**3')
    assert all(a**3 == expected)


def test_less_than():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='s')
    b = osyris.Array(values=[6., 7., 1., 4., 10.], unit='s')
    expected = [True, True, False, False, True]
    np.array_equal(a < b, expected)


def test_less_than_bad_units():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='s')
    b = osyris.Array(values=[6., 7., 1., 4., 10.], unit='m')
    with pytest.raises(pint.errors.DimensionalityError):
        _ = a < b


def test_less_equal():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='s')
    b = osyris.Array(values=[6., 7., 1., 4., 10.], unit='s')
    expected = [True, True, False, True, True]
    np.array_equal(a <= b, expected)


def test_less_equal_bad_units():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='s')
    b = osyris.Array(values=[6., 7., 1., 4., 10.], unit='m')
    with pytest.raises(pint.errors.DimensionalityError):
        _ = a <= b


def test_greater_than():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='s')
    b = osyris.Array(values=[6., 7., 1., 4., 10.], unit='s')
    expected = [True, True, False, False, True]
    np.array_equal(b > a, expected)


def test_greater_than_bad_units():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='s')
    b = osyris.Array(values=[6., 7., 1., 4., 10.], unit='K')
    with pytest.raises(pint.errors.DimensionalityError):
        _ = b > a


def test_greater_equal():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='s')
    b = osyris.Array(values=[6., 7., 1., 4., 10.], unit='s')
    expected = [True, True, False, True, True]
    np.array_equal(b <= a, expected)


def test_greater_equal_bad_units():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='s')
    b = osyris.Array(values=[6., 7., 1., 4., 10.], unit='K')
    with pytest.raises(pint.errors.DimensionalityError):
        _ = b >= a


def test_to():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[1.0e-3, 2.0e-3, 3.0e-3, 4.0e-3, 5.0e-3], unit='km')
    assert all(a.to('km') == b)


def test_to_bad_units():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    with pytest.raises(pint.errors.DimensionalityError):
        _ = a.to('s')


def test_min():
    a = osyris.Array(values=[1., 2., 0.3, 4., 5.], unit='m')
    b = osyris.Array(values=[0.3], unit='m')
    assert all(a.min() == b)


def test_max():
    a = osyris.Array(values=[1., 2., 3., 45., 5.], unit='m')
    b = osyris.Array(values=[45.0], unit='m')
    assert all(a.max() == b)


def test_reshape():
    a = osyris.Array(values=[1., 2., 3., 4., 5., 6.], unit='m')
    expected = osyris.Array(values=[[1., 2., 3.], [4., 5., 6.]], unit='m')
    assert all(np.ravel(a.reshape(2, 3) == expected))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

from .map import map as _map
import warnings


def plane(*args, **kwargs):
    """
    Old deprecated alias for think map, will be removed soon.
    """
    warnings.warn("The plane function is deprecated and will be removed soon, "
                  "use map instead.")
    return _map(*args, dz=None, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)

# flake8: noqa

import sys
import struct
import numpy as np


def to_vtk(holder, fname="osyris_data.vtu", dx=None, dy=None, dz=None):
    """
    Write RAMSES data to VTK file
    """

    try:
        from scipy.spatial import Delaunay
    except ImportError:
        print(
            "Scipy Delaunay library not found. This is needed for VTK output. Exiting.")

    # Print status
    if not fname.endswith(".vtu"):
        fname += ".vtu"
    print("Writing data to VTK file: " + fname)

    if dx is not None:
        if holder.info["ndim"] == 1:
            cube = np.where(np.logical_and(holder.get("x")>=-0.5*dx, \
                                           holder.get("x")<= 0.5*dx))
        elif holder.info["ndim"] == 2:
            cube = np.where(np.logical_and(holder.get("x")>=-0.5*dx, \
                            np.logical_and(holder.get("x")<= 0.5*dx, \
                            np.logical_and(holder.get("y")>=-0.5*dy, \
                                           holder.get("y")<= 0.5*dy))))
        elif holder.info["ndim"] == 3:
            cube = np.where(np.logical_and(holder.get("x")>=-0.5*dx, \
                            np.logical_and(holder.get("x")<= 0.5*dx, \
                            np.logical_and(holder.get("y")>=-0.5*dy, \
                            np.logical_and(holder.get("y")<= 0.5*dy, \
                            np.logical_and(holder.get("z")>=-0.5*dz, \
                                           holder.get("z")<= 0.5*dz))))))
        points = np.array(
            [holder.get("x")[cube],
             holder.get("y")[cube],
             holder.get("z")[cube]]).T
    else:
        # Coordinates ot RAMSES cell centers
        points = np.array([holder.get("x"), holder.get("y"), holder.get("z")]).T

    # Compute Delaunay tetrahedralization from cell nodes
    # Note that this step can take a lot of time!
    ncells = points.shape[0]
    print("Computing Delaunay mesh with %i points." % ncells)
    print("This may take some time...")
    tri = Delaunay(points, qhull_options="QJ Qx Qs Qv")
    ntetra = np.shape(tri.simplices)[0]
    nverts = ntetra * 4
    print("Delaunay mesh with %i tetrahedra complete." % ntetra)

    # Create list of variables by grouping x,y,z components together
    key_list = holder.get_var_list()
    n_components = []
    varlist = []
    for key in key_list:
        thisVar = getattr(holder, key)
        if (not thisVar.vector_component) and (thisVar.group !=
                                               "amr") and (thisVar.group != "part"):
            if thisVar.kind == "vector":
                n_components.append(3)
            elif thisVar.kind == "scalar":
                n_components.append(1)
            else:
                print("Unknown data type: " + thisVar.kind)
                return
            varlist.append(key)

    nvars = len(n_components)

    # Compute byte sizes
    nbytes_xyz = 3 * ncells * 8
    nbytes_cellc = nverts * 4
    nbytes_cello = ntetra * 4
    nbytes_cellt = ntetra * 4
    nbytes_vars = np.zeros([nvars], dtype=np.int32)
    for i in range(nvars):
        nbytes_vars[i] = n_components[i] * ncells * 8

    # Compute byte offsets
    offsets = np.zeros([nvars + 4], dtype=np.int64)
    offsets[0] = 0  # xyz coordinates
    offsets[1] = offsets[0] + 4 + nbytes_xyz  # cell connectivity
    offsets[2] = offsets[1] + 4 + nbytes_cellc  # cell offsets
    offsets[3] = offsets[2] + 4 + nbytes_cello  # cell types
    offsets[4] = offsets[3] + 4 + nbytes_cellt  # first hydro variable
    for i in range(nvars - 1):
        offsets[i + 5] = offsets[i + 4] + 4 + nbytes_vars[i]

    # Open file for binary output
    with open(fname, "wb") as f:
        # Write VTK file header
        f.write(string_to_binary('<?xml version=\"1.0\"?>'))
        f.write(
            string_to_binary(
                '<VTKFile type=\"UnstructuredGrid\" version=\"0.1\" byte_order=\"LittleEndian\">'
            ))
        f.write(string_to_binary('   <UnstructuredGrid>'))
        f.write(
            string_to_binary('   <Piece NumberOfPoints=\"%i\" NumberOfCells=\"%i\">' %
                             (ncells, ntetra)))
        f.write(string_to_binary('      <Points>'))
        f.write(
            string_to_binary(
                '         <DataArray type=\"Float64\" Name=\"Coordinates\" NumberOfComponents=\"3\" format=\"appended\" offset=\"%i\" />'
                % offsets[0]))
        f.write(string_to_binary('      </Points>'))
        f.write(string_to_binary('      <Cells>'))
        f.write(
            string_to_binary(
                '         <DataArray type=\"Int32\" Name=\"connectivity\" format=\"appended\" offset=\"%i\" />'
                % offsets[1]))
        f.write(
            string_to_binary(
                '         <DataArray type=\"Int32\" Name=\"offsets\" format=\"appended\" offset=\"%i\" />'
                % offsets[2]))
        f.write(
            string_to_binary(
                '         <DataArray type=\"Int32\" Name=\"types\" format=\"appended\" offset=\"%i\" />'
                % offsets[3]))
        f.write(string_to_binary('      </Cells>'))
        f.write(string_to_binary('      <PointData>'))
        for i in range(nvars):
            f.write(
                string_to_binary(
                    '         <DataArray type=\"Float64\" Name=\"' + varlist[i] +
                    '\" NumberOfComponents=\"%i\" format=\"appended\" offset=\"%i\" />'
                    % (n_components[i], offsets[i + 4])))
        f.write(string_to_binary('      </PointData>'))
        f.write(string_to_binary('   </Piece>'))
        f.write(string_to_binary('   </UnstructuredGrid>'))
        f.write(string_to_binary('   <AppendedData encoding=\"raw\">'))
        f.write(string_to_binary('_', newline=False))

        # Now write data in binary. Every data field is preceded by its byte size.
        # x,y,z coordinates of the points
        f.write(struct.pack('<i', *[nbytes_xyz]))
        f.write(struct.pack('<%id' % (ncells * 3), *np.ravel(points)))

        # Cell connectivity
        f.write(struct.pack('<i', *[nbytes_cellc]))
        f.write(struct.pack('<%ii' % nverts, *np.ravel(tri.simplices)))

        # Cell offsets
        f.write(struct.pack('<i', *[nbytes_cello]))
        f.write(struct.pack('<%ii' % ntetra, *range(4, ntetra * 4 + 1, 4)))

        # Cell types: number 10 is tetrahedron in VTK file format
        f.write(struct.pack('<i', *[nbytes_cellt]))
        f.write(struct.pack('<%ii' % ntetra, *np.full(ntetra, 10, dtype=np.int32)))

        # Cell variables
        for i in range(nvars):
            if n_components[i] == 3:
                if dx is not None:
                    celldata = np.ravel(
                        np.array([
                            holder.get(varlist[i] + "_x")[cube],
                            holder.get(varlist[i] + "_y")[cube],
                            holder.get(varlist[i] + "_z")[cube]
                        ]).T)
                else:
                    celldata = np.ravel(
                        np.array([
                            holder.get(varlist[i] + "_x"),
                            holder.get(varlist[i] + "_y"),
                            holder.get(varlist[i] + "_z")
                        ]).T)
            else:
                if dx is not None:
                    celldata = holder.get(varlist[i])[cube]
                else:
                    celldata = holder.get(varlist[i])
            f.write(struct.pack('<i', *[nbytes_vars[i]]))
            f.write(struct.pack('<%id' % (ncells * n_components[i]), *celldata))

        # Close file
        f.write(string_to_binary('   </AppendedData>'))
        f.write(string_to_binary('</VTKFile>'))
        # f.close()

    # File size
    fsize_raw = offsets[nvars + 3] + nbytes_vars[nvars - 1]
    if fsize_raw > 1.0e9:
        fsize = float(fsize_raw) / 1.0e9
        funit = "GB"
    elif fsize_raw > 1.0e6:
        fsize = float(fsize_raw) / 1.0e6
        funit = "MB"
    elif fsize_raw > 1000:
        fsize = float(fsize_raw) / 1.0e3
        funit = "KB"
    else:
        fsize = float(fsize_raw)
        funit = "B"

    print("File " + fname + (" of size %.1f" % fsize) + funit + " succesfully written.")

    return


def string_to_binary(s, newline=True):
    if sys.version_info < (3, ):
        return s + ("\n" if newline else "")
    else:
        return str.encode(s) + (b"\n" if newline else b"")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)
import numpy as np
import os
from .reader import Reader, ReaderKind
from ..core import Array
from . import utils


class PartReader(Reader):
    def __init__(self):
        super().__init__(kind=ReaderKind.PART)

    def initialize(self, meta, select):
        self.initialized = False
        if select is False:
            return

        # Read the number of variables from the hydro_file_descriptor.txt
        # and select the ones to be read if specified by user
        fname = os.path.join(meta["infile"], "part_file_descriptor.txt")
        try:
            desc_from_file = np.loadtxt(fname, dtype=str, delimiter=",")
        except IOError:
            return

        descriptor = {
            desc_from_file[i, 1].strip(): desc_from_file[i, 2].strip()
            for i in range(len(desc_from_file))
        }

        self.descriptor_to_variables(descriptor=descriptor, meta=meta, select=select)
        self.initialized = True

    def read_header(self, info):
        if not self.initialized:
            return
        self.offsets["i"] += 2
        self.offsets["n"] += 2
        [nparticles] = utils.read_binary_data(fmt="i",
                                              content=self.bytes,
                                              offsets=self.offsets)
        for i in range(5):
            self.offsets["b"] += utils.skip_binary_line(content=self.bytes,
                                                        offsets=self.offsets)
        for item in self.variables.values():
            if item["read"]:
                npieces = len(item["pieces"])
                item["pieces"][npieces] = Array(values=np.array(
                    utils.read_binary_data(fmt="{}{}".format(nparticles, item["type"]),
                                           content=self.bytes,
                                           offsets=self.offsets)) *
                                                item["unit"].magnitude,
                                                unit=1.0 * item["unit"].units)
            else:
                self.offsets[item["type"]] += nparticles
                self.offsets["n"] += 1
        info["nparticles"] += nparticles

    def allocate_buffers(self, ngridmax, twotondim):
        return

    def read_variables(self, ncache, ind, ilevel, cpuid, info):
        return

    def make_conditions(self, select, ncache):
        return {}

    def read_footer(self, *args, **kwargs):
        return

    def step_over(self, ncache, twotondim, ndim):
        return



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/nvaytet/osyris)
import osyris


def test_dataset_creation():
    a = osyris.Array(values=[1., 2., 3., 4., 5.], unit='m')
    b = osyris.Array(values=[6., 7., 8., 9., 10.], unit='s')
    dg1 = osyris.Datagroup()
    dg1['a'] = a
    dg1['b'] = b
    dg2 = osyris.Datagroup()
    dg2['a'] = a[:-1]
    dg2['b'] = b[:-1]
    ds = osyris.Dataset()
    ds['one'] = dg1
    ds['two'] = dg2
    assert 'one' in ds
    assert 'two' in ds
